; 
    include \masm32\include\masm32rt.inc
    include limoalloc.asm
; 

    ALLOC_MAX EQU 100000h    ; actual max is 4 times this value

; 

alignment MACRO pv
    xor eax, eax
    mov ecx, pv
    bsf ecx, ecx
    jz @F
    mov eax, 1
    shl eax, cl
  @@:
    EXITM <eax>
ENDM

; 

    PROCESS_HEAP_ENTRY STRUCT
      lpData        DWORD ?
      cbData        DWORD ?
      cbOverhead    BYTE  ?
      iRegionIndex  BYTE  ?
      wFlags        WORD  ?
      UNION
        STRUCT
          hMem        DWORD ?
          dwReserved  DWORD 4 dup(?)
        ENDS
        STRUCT
          dwCommittedSize   DWORD ?
          dwUnCommittedSize DWORD ?
          lpFirstBlock      DWORD ?
          lpLastBlock       DWORD ?
        ENDS
      ENDS
    PROCESS_HEAP_ENTRY ENDS

; 
    .data
      nAlloc      dd 0
      nReallocUp  dd 0
      nReallocDn  dd 0
      nFrees      dd 0
      nErrors     dd 0
      allocMax    dd 0
      allocMin    dd ALLOC_MAX
      alignMin    dd ALLOC_MAX

      phe PROCESS_HEAP_ENTRY  <>
    .code
; 

HeapAllocatedTotal proc uses ebx esi
    mov phe.lpData, 0
    invoke GetProcessHeap
    mov ebx, eax
    xor esi, esi
  @@:
    invoke HeapWalk, ebx, ADDR phe
    test eax, eax
    jz  @F
    add esi, phe.cbData
    jmp @B
  @@:
    mov eax, esi
    ret
HeapAllocatedTotal endp

; 

HeapBusyEntries proc uses ebx esi
    mov phe.lpData, 0
    invoke GetProcessHeap
    mov ebx, eax
    xor esi, esi
  @@:
    invoke HeapWalk, ebx, ADDR phe
    test eax, eax
    jz  @F
    test phe.wFlags, PROCESS_HEAP_ENTRY_BUSY
    jz  @B
    inc esi
    jmp @B
  @@:
    mov eax, esi
    ret
HeapBusyEntries endp

; 

HeapEntrySize proc uses ebx edi pv:DWORD

    mov edi, pv
    mov phe.lpData, 0
    invoke GetProcessHeap
    mov ebx, eax
  @@:
    invoke HeapWalk, ebx, ADDR phe
    test eax, eax
    jz  @F
    .IF phe.lpData == edi
      mov eax, phe.cbData
      ret
    .ENDIF
    jmp @B
  @@:
    mov eax, -1
    ret

HeapEntrySize endp

; 

DumpHeap proc uses ebx esi edi lpFilename:DWORD

    mov edi, fcreate(lpFilename)

    mov phe.lpData, 0
    invoke GetProcessHeap
    mov ebx, eax
  @@:
    invoke HeapWalk, ebx, ADDR phe
    test eax, eax
    jz  @F

    fprintc edi,uhex$(phe.lpData)
    fprintc edi,"\n\tcbData      \t"
    fprintc edi,ustr$(phe.cbData)

    fprintc edi,"\n\tcbOverhead  \t"
    movzx eax, phe.cbOverhead
    fprintc edi,ustr$(eax)
    fprintc edi,"\n\tiRegionIndex\t"
    movzx eax, phe.iRegionIndex
    fprintc edi,ustr$(eax)

    fprintc edi,"\n\twFlags\n"
    movzx esi, phe.wFlags
    .IF esi & PROCESS_HEAP_REGION
      fprintc edi,"\t  PROCESS_HEAP_REGION\n"
    .ELSEIF esi & PROCESS_HEAP_UNCOMMITTED_RANGE
      fprintc edi,"\t  PROCESS_HEAP_UNCOMMITTED_RANGE\n"
    .ELSEIF esi & PROCESS_HEAP_ENTRY_BUSY
      fprintc edi,"\t  PROCESS_HEAP_ENTRY_BUSY\n"
    .ELSEIF esi & PROCESS_HEAP_ENTRY_MOVEABLE
      fprintc edi,"\t  PROCESS_HEAP_ENTRY_MOVEABLE\n"
    .ELSEIF esi & PROCESS_HEAP_ENTRY_DDESHARE
      fprintc edi,"\t  PROCESS_HEAP_ENTRY_DDESHARE\n"
    .ENDIF

    .IF esi & PROCESS_HEAP_ENTRY_BUSY && esi & PROCESS_HEAP_ENTRY_MOVEABLE
      fprintc edi,"\thMem        \t"
      fprintc edi,uhex$(phe.hMem)
      fprintc edi,"\n"
    .ENDIF

    .IF esi & PROCESS_HEAP_REGION
      fprintc edi,"\tdwCommittedSize  \t"
      fprintc edi,ustr$(phe.dwCommittedSize)
      fprintc edi,"\n\tdwUnCommittedSize\t"
      fprintc edi,ustr$(phe.dwUnCommittedSize)
      fprintc edi,"\n\tlpFirstBlock     \t"
      fprintc edi,ustr$(phe.lpFirstBlock)
      fprintc edi,"\n\tlpLastBlock      \t"
      fprintc edi,ustr$(phe.lpLastBlock)
    .ENDIF

    fprintc edi,"\n"

    jmp @B
  @@:
    fclose edi
    ret

DumpHeap endp

; 

Probe proc pv:DWORD, cb:DWORD

    mov edx, pv
    mov eax, [edx]
    mov [edx], eax
    add edx, cb
    sub edx, 4
    mov eax, [edx]
    mov [edx], eax
    ret

Probe endp

; 
start:
; 

    invoke DumpHeap, chr$("dump1.txt")

    print "heap busy entries    : "
    invoke HeapBusyEntries
    print ustr$(eax),13,10
    print "heap allocated total : "
    invoke HeapAllocatedTotal
    print ustr$(eax),13,10,13,10
    inkey "press any key to start...",13,10

    invoke nseed, rv(GetTickCount)
  looper:
    invoke nrandom, ALLOC_MAX
    inc eax
    mov ebx, eax
    invoke _Alloc, ebx
    mov esi, eax
    inc nAlloc

    invoke Probe, esi, ebx

    .IF ebx > allocMax
      mov allocMax, ebx
    .ELSEIF ebx < allocMin
      mov allocMin, ebx
    .ENDIF

    .IF alignment(esi) < alignMin
      mov alignMin, eax
    .ENDIF

    invoke HeapEntrySize, esi
    .IF eax < ebx
      inc nErrors
    .ENDIF
    push eax
    invoke GetSize, esi
    pop edx
    .IF eax != edx
      inc nErrors
    .ENDIF

    shl ebx, 2
    invoke Realloc, esi, ebx
    mov esi, eax
    inc nReallocUp

    invoke Probe, esi, ebx

    .IF ebx > allocMax
      mov allocMax, ebx
    .ENDIF

    invoke HeapEntrySize, esi
    .IF eax < ebx
      inc nErrors
    .ENDIF
    push eax
    invoke GetSize, esi
    pop edx
    .IF eax != edx
      inc nErrors
    .ENDIF

    .IF alignment(esi) < alignMin
      mov alignMin, eax
    .ENDIF

    shr ebx, 4
    .IF ebx < 1
      inc ebx
    .ENDIF
    invoke Realloc, esi, ebx
    mov esi, eax
    inc nReallocDn

    invoke Probe, esi, ebx

    invoke HeapEntrySize, esi
    .IF eax < ebx
      inc nErrors
    .ENDIF
    push eax
    invoke GetSize, esi
    pop edx
    .IF eax != edx
      inc nErrors
    .ENDIF

    .IF alignment(esi) < alignMin
      mov alignMin, eax
    .ENDIF

    invoke _Free, esi
    inc nFrees

    invoke HeapEntrySize, esi
    .IF eax != 0ffffffffh
      inc nErrors
    .ENDIF
    movzx eax, phe.wFlags
    .IF eax != PROCESS_HEAP_UNCOMMITTED_RANGE
      inc nErrors
    .ENDIF

    invoke locate, 0, 5
    print "allocations          : "
    print ustr$(nAlloc),13,10
    print "reallocations up     : "
    print ustr$(nReallocUp),13,10
    print "reallocations down   : "
    print ustr$(nReallocDn),13,10
    print "allocation frees     : "
    print ustr$(nFrees),13,10
    print "maximum allocation   : "
    print ustr$(allocMax)," ",13,10
    print "minimum allocation   : "
    print ustr$(allocMin),"         ",13,10
    print "minimum alignment    : "
    print ustr$(alignMin),"         ",13,10

    print "errors               : "
    print ustr$(nErrors),13,10,13,10
    print "press any key to stop...",13,10,13,10

    invoke crt__kbhit

    test eax, eax
    jz looper

    invoke DumpHeap, chr$("dump2.txt")

    print "heap busy entries    : "
    invoke HeapBusyEntries
    print ustr$(eax),13,10
    print "heap allocated total : "
    invoke HeapAllocatedTotal
    print ustr$(eax),13,10,13,10

    inkey "press any key to exit..."
    exit
; 
end start

